package net.knarcraft.ffmpegconverter.utility;

import net.knarcraft.ffmpegconverter.parser.ConverterArgument;
import net.knarcraft.ffmpegconverter.parser.ConverterArgumentValue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * A class to help with command parsing
 */
public final class Parser {

    private Parser() {
    }

    /**
     * This function parses command inputs into understandable converter instructions
     *
     * @param tokens <p>A list of tokens containing all arguments</p>
     */
    private static void parse(List<String> tokens) {
        String[] types = {"animeconverter", "audioconverter", "videoconverter"};
        ConverterArgument[] commonArgs = {
                new ConverterArgument("recursions", 'r', true, ConverterArgumentValue.INT),
                new ConverterArgument("infile", 'i', true, ConverterArgumentValue.STRING)
        };
        ConverterArgument[] animeArgs = {
                new ConverterArgument("audiolang", 'a', true, ConverterArgumentValue.COMMA_SEPARATED_LIST),
                new ConverterArgument("subtitlelang", 's', true, ConverterArgumentValue.COMMA_SEPARATED_LIST),
                new ConverterArgument("tostereo", 't', false, ConverterArgumentValue.BOOLEAN),
                new ConverterArgument("preventpartialsubtitles", 'p', false, ConverterArgumentValue.BOOLEAN)
        };
        ConverterArgument[] audioVideoArgs = {
                new ConverterArgument("-outext", 'o', true,
                        ConverterArgumentValue.STRING)
        };
        String type = tokens.remove(0).toLowerCase();
        if (!ListUtil.listContains(types, s -> s.equals(type))) {
            throw new IllegalArgumentException("Unknown converter type chosen.");
        }
        if (tokens.size() < 2) {
            throw new IllegalArgumentException("No file/folder path in argument.");
        }
        List<ConverterArgument> converterArguments;
        switch (type) {
            case "animeconverter":
                converterArguments = new ArrayList<>(Arrays.asList(ListUtil.concatenate(commonArgs, animeArgs)));
                break;
            case "audioconverter":
            case "videoconverter":
                converterArguments = new ArrayList<>(Arrays.asList(ListUtil.concatenate(commonArgs, audioVideoArgs)));
                break;
            default:
                throw new IllegalArgumentException("Unknown converter type chosen.");
        }
        while (!tokens.isEmpty()) {
            String currentToken = tokens.remove(0);
            List<ConverterArgument> foundArguments;
            if (currentToken.startsWith("--")) {
                String argumentName = currentToken.substring(2);
                foundArguments = ListUtil.getMatching(converterArguments, (item) -> item.getName().equals(argumentName));
            } else if (currentToken.startsWith("-")) {
                String argumentShorthand = currentToken.substring(1);
                foundArguments = ListUtil.getMatching(converterArguments, (item) -> item.getName().equals(argumentShorthand));
            } else {
                throw new IllegalArgumentException("Unexpected value when not given an argument.");
            }
            if (foundArguments.isEmpty()) {
                throw new IllegalArgumentException(String.format("Invalid argument %s encountered.", currentToken));
            }
            ConverterArgument foundArgument = foundArguments.get(0);
            String argumentValue = tokens.get(0);
            //TODO: Check if the value is an argument, and if not consume the value
            //TODO: If the value is an argument, but a value is necessary, throw an error
            //TODO: Store the found parameter values into a hashmap and send it to the converter as input
        }
    }

    /**
     * Tokenizes a string
     *
     * @param input <p>A string.</p>
     * @return <p>A list of tokens.</p>
     */
    public static List<String> tokenize(String input) {
        List<String> tokens = new ArrayList<>();
        boolean startedQuote = false;
        StringBuilder currentToken = new StringBuilder();
        for (int index = 0; index < input.length(); index++) {
            char character = input.charAt(index);
            switch (character) {
                case ' ':
                    if (tokenizeSpace(startedQuote, currentToken, tokens)) {
                        currentToken = new StringBuilder();
                    }
                    break;
                case '"':
                    if (startedQuote) {
                        //This quote signifies the end of the argument
                        if (isNotEmpty(currentToken)) {
                            tokens.add(currentToken.toString());
                            currentToken = new StringBuilder();
                        }
                        startedQuote = false;
                    } else {
                        //This quote signifies the start of the argument
                        startedQuote = true;
                        currentToken = new StringBuilder();
                    }
                    break;
                default:
                    tokenizeNormalCharacter(currentToken, character, input.length(), index, tokens);
                    break;
            }
        }
        return tokens;
    }

    /**
     * Adds a normal character to the token. Adds the current token to tokens if at the end of the input
     *
     * @param currentToken <p>The string builder containing the current token.</p>
     * @param character    <p>The character found in the input.</p>
     * @param inputLength  <p>The length of the given input</p>
     * @param index        <p>The index of the read character.</p>
     * @param tokens       <p>The list of processed tokens.</p>
     */
    private static void tokenizeNormalCharacter(StringBuilder currentToken, char character, int inputLength, int index,
                                                List<String> tokens) {
        currentToken.append(character);
        if (index == inputLength - 1) {
            tokens.add(currentToken.toString());
        }
    }

    /**
     * Tokenizes a space character
     *
     * @param startedQuote <p>Whether this space is inside a pair of quotes.</p>
     * @param currentToken <p>The string builder containing the current token.</p>
     * @param tokens       <p>The list of processed tokens.</p>
     * @return <p>True if the token is finished.</p>
     */
    private static boolean tokenizeSpace(boolean startedQuote, StringBuilder currentToken, List<String> tokens) {
        if (!startedQuote) {
            //If not inside "", a space marks the end of a parameter
            if (isNotEmpty(currentToken)) {
                tokens.add(currentToken.toString());
            }
            return true;
        } else {
            currentToken.append(' ');
            return false;
        }
    }

    /**
     * Checks whether a string builder is empty
     *
     * @param builder <p>The string builder to check.</p>
     * @return <p>True if the string builder is non empty.</p>
     */
    private static boolean isNotEmpty(StringBuilder builder) {
        return !builder.toString().trim().equals("");
    }
}
